from nl4dv.utils import constants, helpers
import copy
import nltk
from nltk import word_tokenize


class AttributeGenie:
    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance

    # Get a sorted list of (attributes and their datatype) tuples to determine IMPLICIT TASKS as well as default VIS encodings
    def get_attr_datatype_shorthand(self, attributes):
        # Attribute-Datatype pair
        unsorted_attr_datatype = [(attr, self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]['dataType']) for attr in attributes]

        # Since the `vis_combo` mapping keys are in a specific order [Q,N,O,T], we will order the list of attributes in this order
        default_sort_order = ['Q', 'N', 'O', 'T']
        sorted_attr_datatype = [(attr,attr_type) for x in default_sort_order for (attr,attr_type) in unsorted_attr_datatype if attr_type == x]

        sorted_attributes = [x[0] for x in sorted_attr_datatype]  # e.g. ['Rotten Tomatoes Rating', 'Worldwide Gross']
        sorted_attribute_datatypes = ''.join([x[1] for x in sorted_attr_datatype])  # e.g. 'QQ'

        return sorted_attributes, sorted_attribute_datatypes

    # 1-time generation of Attributes for processing
    def get_data_attributes(self):
        attribute_map = dict()
        for attr in self.nl4dv_instance.data_genie_instance.data_attribute_map:
            attr_lower = attr.lower()
            attribute_map[attr] = dict()
            attribute_map[attr]['raw'] = attr
            attribute_map[attr]['lower'] = attr_lower
            attribute_map[attr]['stemmed_lower'] = ' '.join(self.nl4dv_instance.porter_stemmer_instance.stem(t) for t in nltk.word_tokenize(attr_lower))

        return attribute_map

    # 1-time generation of Attribute Aliases for processing
    def get_attribute_aliases(self):
        attribute_aliases = dict()
        for attr in self.nl4dv_instance.data_genie_instance.data_attribute_map:
            attribute_aliases[attr] = dict()
            for alias in self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]['aliases']:
                alias_lower = alias.lower()

                attribute_aliases[attr][alias] = dict()
                attribute_aliases[attr][alias]['raw'] = alias
                attribute_aliases[attr][alias]['lower'] = alias_lower
                attribute_aliases[attr][alias]['stemmed_lower'] = ' '.join(self.nl4dv_instance.porter_stemmer_instance.stem(t) for t in nltk.word_tokenize(alias_lower))

        return attribute_aliases

    # Update the state if keyword-attribute mappings
    def update_keyword_attribute_mappings(self, keyword, attribute, score):
        if keyword not in self.nl4dv_instance.keyword_attribute_mapping:
            self.nl4dv_instance.keyword_attribute_mapping[keyword] = dict()
        self.nl4dv_instance.keyword_attribute_mapping[keyword][attribute] = score

        if attribute not in self.nl4dv_instance.attribute_keyword_mapping:
            self.nl4dv_instance.attribute_keyword_mapping[attribute] = dict()
        self.nl4dv_instance.attribute_keyword_mapping[attribute][keyword] = score

    # Detect attributes by exact_match of n-grams
    def detect_attributes_by_exact_match(self, query_ngrams, data_attributes, query_attributes):
        for attr in data_attributes:
            for ngram in query_ngrams:
                # Exact Match
                if data_attributes[attr]["lower"] == query_ngrams[ngram]["lower"] or data_attributes[attr]["stemmed_lower"] == query_ngrams[ngram]["stemmed_lower"]:
                    if attr not in query_attributes:
                        query_attributes[attr] = {
                            'name': attr,
                            "queryPhrase": [query_ngrams[ngram]["lower"]],
                            'inferenceType': 'explicit',
                            'matchScore': self.nl4dv_instance.match_scores['attribute']['attribute_exact_match'],
                            'metric': ['attribute_exact_match'],
                            'isLabel': self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                            'encode': not self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                            'isAmbiguous': False,
                            'ambiguity': list(),
                            'meta': {
                                'score': 100,
                                'threshold': self.nl4dv_instance.match_thresholds['string_similarity'],
                                'alias': None,
                                'ambiguity': {}
                            }
                        }

                        # Update Keyword-Attribute-Score Mappings
                        self.update_keyword_attribute_mappings(keyword=query_ngrams[ngram]["lower"], attribute=attr, score=query_attributes[attr]["matchScore"])

                        # Important! If, the attribute is detected by exact match, then break the loop. We can safely SKIP subsequent n-grams saving *some* cycles.
                        break

        return query_attributes

    # Detect attributes based on string similarity of the n-grams
    def detect_attributes_by_similarity(self, query_ngrams, data_attributes, query_attributes):

        for attr in data_attributes:

            # If already found, continue
            if attr in query_attributes and any(a for a in query_attributes[attr]["metric"] if a in ["attribute_exact_match"]):
                continue

            for ngram in query_ngrams:
                score = 0
                add_attribute = False
                # Compute similarity of tokens
                string_similarity_score = helpers.compute_similarity(data_attributes[attr]["lower"], query_ngrams[ngram]["lower"], 'token_similarity')
                if string_similarity_score >= self.nl4dv_instance.match_thresholds['string_similarity']:
                    score = string_similarity_score
                    add_attribute = True

                else:
                    # Compute similarity of stemmed tokens
                    stemmed_string_similarity_score = helpers.compute_similarity(data_attributes[attr]["stemmed_lower"], query_ngrams[ngram]["stemmed_lower"], 'token_similarity')
                    if stemmed_string_similarity_score >= self.nl4dv_instance.match_thresholds['string_similarity']:
                        score = stemmed_string_similarity_score
                        add_attribute = True

                if add_attribute:
                    if attr not in query_attributes or query_attributes[attr]["meta"]["score"] <= score:
                        query_attributes[attr] = {
                            'name': attr,
                            "queryPhrase": [query_ngrams[ngram]["lower"]],
                            'inferenceType': 'explicit',
                            'matchScore': self.nl4dv_instance.match_scores['attribute']['attribute_similarity_match'],
                            'metric': ['attribute_similarity_match'],
                            'isLabel': self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                            'encode': not self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                            'isAmbiguous': False,
                            'ambiguity': list(),
                            'meta': {
                                'score': score,
                                'threshold': self.nl4dv_instance.match_thresholds['string_similarity'],
                                'alias': None,
                                'ambiguity': {}
                            }
                        }

                        # Update Keyword-Attribute-Score Mappings
                        self.update_keyword_attribute_mappings(keyword=query_ngrams[ngram]["lower"], attribute=attr, score=query_attributes[attr]["matchScore"])

        return query_attributes

    # Detect attributes by alias exact_match of n-grams
    def detect_attributes_by_alias_exact_match(self, query_ngrams, data_attributes, query_attributes, attribute_aliases):
        for attr in data_attributes:
            # If already found, continue
            if attr in query_attributes and any(a for a in query_attributes[attr]["metric"] if a in ["attribute_similarity_match"]):
                continue

            for alias in attribute_aliases[attr]:
                is_exact_match = False
                for ngram in query_ngrams:
                    is_exact_match = False
                    if query_ngrams[ngram]["lower"] == attribute_aliases[attr][alias]["lower"] or query_ngrams[ngram]["stemmed_lower"] == attribute_aliases[attr][alias]["stemmed_lower"]:
                        is_exact_match = True
                        if attr not in query_attributes:
                            query_attributes[attr] = {
                                'name': attr,
                                "queryPhrase": [query_ngrams[ngram]["lower"]],
                                'inferenceType': 'explicit',
                                'matchScore': self.nl4dv_instance.match_scores['attribute']['attribute_alias_exact_match'],
                                'metric': 'attribute_alias_exact_match',
                                'isLabel': self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                                'encode': not self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                                'isAmbiguous': False,
                                'ambiguity': list(),
                                'meta': {
                                    'score': 100,
                                    'threshold': self.nl4dv_instance.match_thresholds['string_similarity'],
                                    'alias': alias,
                                    'ambiguity': {}
                                }
                            }

                            # Update Keyword-Attribute-Score Mappings
                            self.update_keyword_attribute_mappings(keyword=query_ngrams[ngram]["lower"], attribute=attr, score=query_attributes[attr]["matchScore"])

                        if is_exact_match:
                            # Important! If, the attribute is detected by exact match, then break the loop. We can safely SKIP subsequent n-grams saving *some* cycles.
                            break

                if is_exact_match:
                    # Important! If, the attribute is detected by exact match, then break the loop. We can safely SKIP subsequent n-grams saving *some* cycles.
                    break

        return query_attributes

    # Detect attributes based on string similarity of the n-grams with the developer specified aliases
    def detect_attributes_by_alias_similarity(self, query_ngrams, data_attributes, query_attributes, attribute_aliases):
 
        # Go over each ngram to check for matches in attribute aliases
        # Check if the word is in one of the attribute aliases
        for attr in data_attributes:

            # If already found, continue
            if attr in query_attributes and any(a for a in query_attributes[attr]["metric"] if a in ["attribute_alias_exact_match"]):
                continue

            for alias in attribute_aliases[attr]:
                for ngram in query_ngrams:
                    score = 0
                    add_attribute = False
                    # Compute similarity
                    string_similarity_score = helpers.compute_similarity(query_ngrams[ngram]["lower"], data_attributes[attr]["lower"], 'token_similarity')
                    if string_similarity_score == 100:
                        add_attribute = True
                        score = string_similarity_score
                    else:
                        # Compute similarity
                        stemmed_string_similarity_score = helpers.compute_similarity(query_ngrams[ngram]["stemmed_lower"], data_attributes[attr]["stemmed_lower"], 'token_similarity')
                        if stemmed_string_similarity_score == 100:
                            add_attribute = True
                            score = stemmed_string_similarity_score

                    if add_attribute:
                        if attr not in query_attributes:
                            query_attributes[attr] = {
                                'name': attr,
                                "queryPhrase": [query_ngrams[ngram]["lower"]],
                                'inferenceType': 'explicit',
                                'matchScore': self.nl4dv_instance.match_scores['attribute']['attribute_alias_similarity_match'],
                                'metric': 'attribute_alias_similarity_match',
                                'isLabel': self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                                'encode': not self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                                'isAmbiguous': False,
                                'ambiguity': list(),
                                'meta': {
                                    'score': score,
                                    'threshold': self.nl4dv_instance.match_thresholds['string_similarity'],
                                    'alias': alias,
                                    'ambiguity': {}
                                }
                            }

                            # Update Keyword-Attribute-Score Mappings
                            self.update_keyword_attribute_mappings(keyword=query_ngrams[ngram]["lower"], attribute=attr, score=query_attributes[attr]["matchScore"])


        return query_attributes

    # Detect attributes based on semantic similarity of the n-grams
    def detect_attributes_by_synonymity(self, query_ngrams, data_attributes, query_attributes):

        # Go over each ngram to check for matches in attribute aliases
        # Check if the word is in one of the attribute aliases

        for attr in data_attributes:

            # If already found, continue
            if attr in query_attributes and any(a for a in query_attributes[attr]["metric"] if a in ["attribute_exact_match","attribute_alias_exact_match", "attribute_similarity_match", "attribute_alias_similarity_match"]):
                continue

            for ngram in query_ngrams:

                # NOTE: Trying to compute wordnet similarity (synonymity) for STEMMED words can be error prone.
                synonym_match_score = helpers.synonymity_score(query_ngrams[ngram]["lower"], data_attributes[attr]["lower"])
                if synonym_match_score >= self.nl4dv_instance.match_thresholds['synonymity']:
                    query_attributes[attr] = {
                        'name': attr,
                        "queryPhrase": [query_ngrams[ngram]["lower"]],
                        'matchScore': self.nl4dv_instance.match_scores['attribute']['attribute_synonym_match'],
                        'inferenceType': 'explicit',
                        'metric': ['attribute_synonym_match'],
                        'isLabel': self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                        'encode': not self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                        'isAmbiguous': False,
                        'ambiguity': list(),
                        'meta': {
                            'score': synonym_match_score,
                            'threshold': self.nl4dv_instance.match_thresholds['synonymity'],
                            'alias': None,
                            'ambiguity': {}
                        }
                    }

                    # Update Keyword-Attribute-Score Mappings
                    self.update_keyword_attribute_mappings(keyword=query_ngrams[ngram]["lower"], attribute=attr, score=query_attributes[attr]["matchScore"])

        return query_attributes

    # Detect attributes based on string similarity of the n-grams with the domain values
    def detect_attributes_from_domain_value(self, query_ngrams, data_attributes, query_attributes):

        value_keyword_mapping = dict()
        keyword_value_mapping = dict()

        for attr in data_attributes:

            # NL4DV does not look for domain value matches in the Label Attribute. Controversial, but that's how we've designed this.
            # Update: Since the addition of two Checks below (similarity score and number_of_words matched) for a domain value match, this is NOT required.
            # if attr == self.nl4dv_instance.label_attribute:
            #     continue

            # Look for domain value matches ONLY for ordinal and nominal variables.
            # For timeseries and quantitative  attribute types, it is difficult to map numbers to attributes AND this is computationally inefficient due to their domain size.
            if self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["dataType"] in ['Q','T']:
                continue

            # RESET for each Attribute
            value_keyword_mapping[attr] = dict()
            keyword_value_mapping[attr] = dict()
            
            for ngram in query_ngrams:
                # We'll use the lower-case attributes
                ngram_str = query_ngrams[ngram]["lower"]

                add_attribute = False
                for d in self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]['domain']:
                    value_raw = str(d)
                    value = value_raw.lower()

                    # Exact match
                    if ngram_str == value:
                        # Value - Keyword
                        value_keyword_mapping[attr][value_raw] = ngram_str

                        # Keyword - Value
                        if ngram_str not in keyword_value_mapping[attr]:
                            keyword_value_mapping[attr][ngram_str] = set()

                        keyword_value_mapping[attr][ngram_str].add(value_raw)
                        add_attribute = True

                    else:

                        string_similarity_score = helpers.compute_similarity(ngram_str, value,'token_similarity')

                        # Check 1: Token Similarity score should be 100, i.e. at least 1 word/n-gram in the query must match the attribute domain value
                        if string_similarity_score == 100:

                            # [OLD] Check 2: The matched n-gram must contain 2 or more words/tokens.
                            # if len(ngram_tokens) >= 2:

                            # Check 2: The matched attribute domain value must either be of length >= 2, i.e. 2 words OR be 1 of 2 possible words.
                            value_tokens = list(word_tokenize(value))
                            ngram_tokens = list(word_tokenize(ngram_str))
                            if len(ngram_tokens) >= 2 or (len(ngram_tokens) == 1 and len(value_tokens) == 2):
                                # Value - Keyword
                                value_keyword_mapping[attr][value_raw] = ngram_str

                                # Keyword - Value
                                if len(ngram_str.split()) <= len(value.split()):
                                    if ngram_str not in keyword_value_mapping[attr]:
                                        keyword_value_mapping[attr][ngram_str] = set()
                                    keyword_value_mapping[attr][ngram_str].add(value_raw)

                                add_attribute = True

                if add_attribute:
                    # Required: To filter out keyword subsets that point to the same attribute, e.g. science fiction, fiction, science
                    for k1 in keyword_value_mapping[attr].copy():
                            for k2 in keyword_value_mapping[attr].copy():
                                if k1!=k2 and k1 in k2:
                                    if k1 in keyword_value_mapping[attr]:
                                        del keyword_value_mapping[attr][k1]

                    # When attributes are double defined
                    metrics = ["attribute_domain_value_match"]
                    if attr in query_attributes:
                        # Update its metric
                        metrics = query_attributes[attr]["metric"]
                        if "attribute_domain_value_match" not in query_attributes[attr]["metric"]:
                            metrics.append("attribute_domain_value_match")

                    query_attributes[attr] = {
                        'name': attr,
                        "queryPhrase": list(keyword_value_mapping[attr].keys()),
                        'inferenceType': 'implicit',
                        'matchScore': self.nl4dv_instance.match_scores['attribute']['attribute_domain_value_match'],
                        'metric': metrics,
                        'isLabel': self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["isLabelAttribute"],
                        'isAmbiguous': False,
                        'ambiguity': list(),
                        'encode': False,
                        'meta': {
                            'score': None,
                            'threshold': None,
                            'alias': None,
                            'ambiguity': {}
                        }
                    }

                    op = dict()
                    for k,v in keyword_value_mapping[attr].items():
                        op[k] = list(v)
                    query_attributes[attr]["meta"]['ambiguity'] = op

                    # Update Keyword-Attribute-Score Mappings
                    self.update_keyword_attribute_mappings(keyword=ngram_str, attribute=attr, score=query_attributes[attr]["matchScore"])

        return query_attributes

    # Main function to extract attributes from the query (query_ngrams to be specific).
    def extract_attributes(self, query_ngrams):
        """
        Return relevant attributes of query

        """
        # Values to be returned
        query_attributes = dict()

        # Map between attribute and (score, corresponding word) from the query
        self.nl4dv_instance.attribute_keyword_mapping = dict()

        # Map between keyword and the attribute to find ambiguous attributes
        self.nl4dv_instance.keyword_attribute_mapping = dict()

        # map of attributes, and their variants - stemmed, lowercase, ...
        data_attributes = self.get_data_attributes()

        # map of attributes and their variants - stemmed, lowercase, ...
        attribute_aliases = self.get_attribute_aliases()

        # Detect attributes by token exact match
        query_attributes = self.detect_attributes_by_exact_match(query_ngrams, data_attributes, query_attributes)

        # Detect attributes by token similarity
        query_attributes = self.detect_attributes_by_similarity(query_ngrams, data_attributes, query_attributes)

        # Detect attributes by alias exact match
        query_attributes = self.detect_attributes_by_alias_exact_match(query_ngrams, data_attributes, query_attributes, attribute_aliases)

        # Detect attributes by alias similarity
        query_attributes = self.detect_attributes_by_alias_similarity(query_ngrams, data_attributes, query_attributes, attribute_aliases)

        # Detect attributes by synonymity
        query_attributes = self.detect_attributes_by_synonymity(query_ngrams, data_attributes, query_attributes)

        # Detect attributes by domain value match
        query_attributes = self.detect_attributes_from_domain_value(query_ngrams, data_attributes, query_attributes)

        # ---------------------------------------------------------------------------------------------------
        # Rule Based Filter to ensure 1 keyword maps to the best attribute(s). THIS IS BY KEYWORD AND BY SCORE
        # ---------
        # Need the one with higher score from 2 attributes selected by same keyword.
        # For eg. Querying 'date' in airplane_crashes dataset results in 'Date'  (attribute similarity match) as well as 'Summary' (attribute domain value match).
        # Choose Date and discard Summary.
        # ---------
        # If same score, retain both.
        # For eg. Querying 'expensive' in cars dataset results in 'Retail Price'  (attribute alias similarity match) as well as 'Dealer Cost' (attribute alias similarity match).
        # Retain both Retail Price and Dealer Cost
        # ---------
        attributes_to_delete = set()
        used_keyword_attribute_mapping = dict()

        for attr in query_attributes:
            keywords = query_attributes[attr]["queryPhrase"]
            score = query_attributes[attr]['matchScore']
            for keyword in keywords:
                if keyword in self.nl4dv_instance.keyword_attribute_mapping:
                    used_keyword_attribute_mapping[keyword] = self.nl4dv_instance.keyword_attribute_mapping[keyword]
                for _attr in self.nl4dv_instance.keyword_attribute_mapping[keyword]:
                    if score > self.nl4dv_instance.keyword_attribute_mapping[keyword][_attr]:
                        attributes_to_delete.add(_attr)
                    elif score < self.nl4dv_instance.keyword_attribute_mapping[keyword][_attr]:
                        attributes_to_delete.add(attr)

        # Delete unused keywords from the main self.nl4dv_instance.keyword_attribute_mapping dictionary
        copy_keyword_attribute_mapping = copy.deepcopy(self.nl4dv_instance.keyword_attribute_mapping)
        for key in copy_keyword_attribute_mapping:
            if key not in used_keyword_attribute_mapping:
                del self.nl4dv_instance.keyword_attribute_mapping[key]

        # Now, again delete a few attributes if they are coming from different keywords. Ensure 1 keyword contributes to 1 attribute
        copy_keyword_attribute_mapping = copy.deepcopy(self.nl4dv_instance.keyword_attribute_mapping)
        for _key in copy_keyword_attribute_mapping:
            for _attr in copy_keyword_attribute_mapping[_key]:
                if _key not in query_attributes[_attr]["queryPhrase"]:
                    del self.nl4dv_instance.keyword_attribute_mapping[_key][_attr]
        # ---------------------------------------------------------------------------------------------------
        # ---------------------------------------------------------------------------------------------------
        # If a keyword is a subset of another keyword
        # DISCARD the attributes with the smaller keyword
        # For e.g, "highway miles per gallon" should select only "highway miles per gallon" and not "city miles per gallon" (due to "miles per gallon")
        copy_keyword_attribute_mapping = copy.deepcopy(self.nl4dv_instance.keyword_attribute_mapping)
        for k1 in copy_keyword_attribute_mapping:
            for k2 in copy_keyword_attribute_mapping:
                if k1 != k2 and k1 in k2:
                    # Remove the "smaller" keyword (e.g. science fiction v/s fiction)
                    if k1 in self.nl4dv_instance.keyword_attribute_mapping:
                        del self.nl4dv_instance.keyword_attribute_mapping[k1]

                    # Remove the attribute of the "smaller" keyword IF it is different from the "bigger" one.
                    for _attr in copy_keyword_attribute_mapping[k1]:
                        if _attr not in copy_keyword_attribute_mapping[k2]:
                            attributes_to_delete.add(_attr)
        # ---------------------------------------------------------------------------------------------------
        # ---------------------------------------------------------------------------------------------------
        #  Delete the now-unwanted attributes and keywords mapped to these attributes.
        for attr in attributes_to_delete:
            if attr in query_attributes:
                # 1) delete unwanted attributes from the main attributes object
                del query_attributes[attr]

                # 2) delete unwanted attributes from the attribute-keyword mapping object
                del self.nl4dv_instance.attribute_keyword_mapping[attr]

                # 3) delete unwanted attributes from the keyword-attributes mapping object
                for k in self.nl4dv_instance.keyword_attribute_mapping:
                    if attr in self.nl4dv_instance.keyword_attribute_mapping[k]:
                       del self.nl4dv_instance.keyword_attribute_mapping[k][attr]

        # Delete unwanted keywords in the finalized attributes
        copy_attribute_keyword_mapping = copy.deepcopy(self.nl4dv_instance.attribute_keyword_mapping)
        for attr in copy_attribute_keyword_mapping:
            for k in copy_attribute_keyword_mapping[attr]:
                if k not in self.nl4dv_instance.keyword_attribute_mapping:
                    del self.nl4dv_instance.attribute_keyword_mapping[attr][k]

        # ---------------------------------------------------------------------------------------------------
        # ---------------------------------------------------------------------------------------------------
        #  Mark attributes as AMBIGUOUS OR NOT. If Ambiguous, append the ambiguities
        for attr in query_attributes:
            # Iterate over it's keywords
            keywords = query_attributes[attr]["queryPhrase"]
            for keyword in keywords:
                if len(self.nl4dv_instance.keyword_attribute_mapping[keyword].keys()) > 1:
                    for ambiguous_attr in self.nl4dv_instance.keyword_attribute_mapping[keyword]:
                        if 'ambiguity' not in query_attributes[attr]:
                            query_attributes[attr]['ambiguity'] = list()

                        # Mark it as ambiguous
                        query_attributes[attr]['isAmbiguous'] = True

                        if ambiguous_attr not in query_attributes[attr]["ambiguity"] and ambiguous_attr != attr:
                            query_attributes[attr]["ambiguity"].append(ambiguous_attr)

                    # Since ambiguous attributes so far have the same score, we compute Ratio Similarity to disambiguate among them.
                    query_attributes[attr]["meta"]["confidence"] = round(helpers.compute_similarity(attr, keyword, "ratio_similarity"), 3)
                else:
                    # Set as unambiguous by default
                    query_attributes[attr]['isAmbiguous'] = False
                    query_attributes[attr]["ambiguity"] = list()
                    query_attributes[attr]["meta"]["confidence"] = 100

        # Clean-up both attribute and keyword mappings, if they are EMPTY / None.
        copy_attribute_keyword_mapping = copy.deepcopy(self.nl4dv_instance.attribute_keyword_mapping)
        for attr in copy_attribute_keyword_mapping:
            if not self.nl4dv_instance.attribute_keyword_mapping[attr]:
                del self.nl4dv_instance.attribute_keyword_mapping[attr]

        copy_keyword_attribute_mapping = copy.deepcopy(self.nl4dv_instance.keyword_attribute_mapping)
        for keyword in copy_keyword_attribute_mapping:
            if not self.nl4dv_instance.keyword_attribute_mapping[keyword]:
                del self.nl4dv_instance.keyword_attribute_mapping[keyword]

        return query_attributes

    # Return just those attributes that are to be "encoded" into the visualization, i.e. along x, y, size, color, column, row, etc.
    def get_encodeable_attributes(self):
        encodeable_attributes = list()

        # Just Take in those attributes, that are tagged as "ENCODE"
        for attr in self.nl4dv_instance.extracted_attributes.keys():
            if self.nl4dv_instance.extracted_attributes[attr]["encode"]:
                encodeable_attributes.append(attr)

        return encodeable_attributes

    # Update attribute encodings based on tasks. e.g. "gross more than 50M" is a filter and the attribute "Worldwide Gross" need not be encoded into the visualization.
    def update_encodeable_attributes_based_on_tasks(self):
        # REFINE ATTRIBUTES based ON TASKS until now
        # Attributes affected by filter should be excluded from the vis UNLESS there are multiple keyword occurrences in the query
        if "filter" in self.nl4dv_instance.extracted_tasks:
            for task_obj in self.nl4dv_instance.extracted_tasks["filter"]:
                # NOTE: remove attributes that are mapped to Filter tasks. We don't ENCODE them explicitly in the VIS.
                for attr in task_obj['attributes']:
                    # If the attribute is detected from domain value that too only IMPLICITLY (as in was not also EXPLICITLY detected), encode it to False
                    if "attribute_domain_value_match" in self.nl4dv_instance.extracted_attributes[attr]["metric"]:
                        if self.nl4dv_instance.extracted_attributes[attr]["inferenceType"] == 'implicit':
                            self.nl4dv_instance.extracted_attributes[attr]["encode"] = False

                    # If the attribute is part of some FILTER (e.g. budget more than 50) but also exists standalone (e.g. correlate budget and gross for budget more than 50)
                    # Simple heuristic check for now: Check for multiple occurrences of the attribute's keyword in the raw query.
                    else:
                        keyword = list(self.nl4dv_instance.attribute_keyword_mapping[attr].keys())[0]
                        if self.nl4dv_instance.query_processed.count(keyword) == 1:
                            self.nl4dv_instance.extracted_attributes[attr]["encode"] = False

        # If correlation as a task is detected, then ensure attributes are ENCODED, even if it means the FILTER ones.
        for task_name in ["correlation", "distribution", "derived_value", "find_extremum", "trend"]:
            if task_name in self.nl4dv_instance.extracted_tasks:
                for task_obj in self.nl4dv_instance.extracted_tasks[task_name]:
                    if task_obj["inferenceType"] == 'explicit':
                        for attr in task_obj['attributes']:
                            self.nl4dv_instance.extracted_attributes[attr]["encode"] = True

        # Get Encocdeable attributes
        encodeable_attributes = self.get_encodeable_attributes()

        # IF there are NO encodeable attributes and IF there are ONLY FILTER TASKs detected as EXPLICIT AND NO other ENCODEable attribute exist,
        # OR if there is a FIND_EXTREMUM task
        # Then ADD the LABEL attribute.
        add_label_attr = False
        if len(encodeable_attributes) == 0:
            if {'filter'} == self.nl4dv_instance.task_genie_instance.get_explicit_tasks():
                add_label_attr = True
        elif len(encodeable_attributes) == 1:
            if "find_extremum" in self.nl4dv_instance.extracted_tasks:
                add_label_attr = True

        if add_label_attr and self.nl4dv_instance.label_attribute is not None:
            # If label attribute was Explicitly detected in the query, "encode" it to True ELSE add it manually
            if self.nl4dv_instance.label_attribute in self.nl4dv_instance.extracted_attributes and \
                    self.nl4dv_instance.extracted_attributes[self.nl4dv_instance.label_attribute]["inferenceType"] == 'explicit':
                self.nl4dv_instance.extracted_attributes[self.nl4dv_instance.label_attribute]["encode"] = True
            else:
                # Check if there is a task BUT no ENCODABLE attribute is detected. In this case, add the label attribute.
                self.nl4dv_instance.extracted_attributes[self.nl4dv_instance.label_attribute] = {
                    'name': self.nl4dv_instance.label_attribute,
                    "queryPhrase": None,
                    'inferenceType': 'implicit',
                    'matchScore': 0,
                    'metric': ['label_attribute'],
                    'isLabel': True,  # OBVIOUSLY
                    'isAmbiguous': False,
                    'ambiguity': [],
                    'encode': True,  # Set to TRUE
                    'meta': {
                        'score': None,
                        'threshold': None,
                        'alias': None,
                        'ambiguity': {}
                    }
                }

            # Update the attribute-keyword and keyword-attribute mapping dictionaries
            self.nl4dv_instance.attribute_keyword_mapping[self.nl4dv_instance.label_attribute] = {self.nl4dv_instance.label_attribute: 1}
            self.nl4dv_instance.keyword_attribute_mapping[self.nl4dv_instance.label_attribute] = {self.nl4dv_instance.label_attribute: 1}

            # Add the label attribute to the list of encodeable_attributes.
            if self.nl4dv_instance.label_attribute not in encodeable_attributes:
                encodeable_attributes.append(self.nl4dv_instance.label_attribute)

        return encodeable_attributes

    def validate_attr_combo(self, attr_combo, query_phrase, allow_subset=False):
        unique_keywords = set()

        # The combination has already incorporated the to be "encode"d attributes.
        for c in attr_combo:
            unique_keywords.add(','.join(self.nl4dv_instance.attribute_keyword_mapping[c].keys()))

        unique_attrs = set()
        for k in self.nl4dv_instance.keyword_attribute_mapping:
            is_encode = True
            for a in self.nl4dv_instance.keyword_attribute_mapping[k]:
                if a in self.nl4dv_instance.extracted_attributes and not self.nl4dv_instance.extracted_attributes[a]["encode"]:
                    is_encode = False
                    break

            if is_encode:
                if ','.join(self.nl4dv_instance.keyword_attribute_mapping[k].keys()) not in unique_attrs:
                    unique_attrs.add(','.join(self.nl4dv_instance.keyword_attribute_mapping[k].keys()))

        if allow_subset:
            return len(attr_combo) != len(unique_keywords) or len(attr_combo) != len(query_phrase)

        # Ensure each attribute comes from a different keyword for the visualization AND all such attributes detected form the visualization.
        return len(attr_combo) != len(unique_attrs) or (len(attr_combo) != len(unique_keywords))



import re
from nltk.corpus import wordnet as wn
from itertools import product
import math
from collections import Counter, MutableMapping
from fuzzywuzzy import fuzz
import Levenshtein
from datetime import date, datetime
import copy
from nl4dv.utils import constants
import json


WORD = re.compile(r'\w+')


# Includes Repetition
def get_ngrams(input, n):
    input = input.split(' ')
    output = []
    for i in range(len(input)-n+1):
        output.append(input[i:i+n])
    return output


# Removes Duplicate N-Grams
def get_ngrams_without_duplicate(input, n):
    input = input.split(' ')
    output = {}
    for i in range(len(input)-n+1):
        g = ' '.join(input[i:i+n])
        output.setdefault(g, 0)
        output[g] += + 1
        return output


# ToDo:- Explore alternatives to WordNet similarity
def synonymity_score(word_x, word_y):
    # type: (str, str) -> int
    """
    Helper method to find similarity between two words

    """
    sem_1 = wn.synsets(word_x)
    sem_2 = wn.synsets(word_y)
    max_score = 0
    for i, j in list(product(*[sem_1, sem_2])):
        score = i.wup_similarity(j)
        max_score = score if (score is not None and max_score < score) else max_score

    return max_score * 100


def set_default(obj):
    """
    Return a serialized version of set  which is a list.

    """
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()

    if isinstance(obj, set):
        return list(obj)

    raise TypeError ("Type %s not serializable" % type(obj))


def cond_print(string_to_print, debug=True):
    if debug:
        print(string_to_print)


def normalize(vis_list, attr, tasks):
    max_score_by_attr = len(attr)
    max_score_by_tasks = len(tasks)
    max_score_by_vis = 1
    for vis in vis_list:
        vis["normalized_score"] = vis["score"] / (max_score_by_attr + max_score_by_tasks + max_score_by_vis)
    return vis_list


def vectorize(text):
    words = WORD.findall(text)
    return Counter(words)


def cosine_similarity(vec1, vec2):
    intersection = set(vec1.keys()) & set(vec2.keys())
    numerator = sum([vec1[x] * vec2[x] for x in intersection])

    sum1 = sum([vec1[x]**2 for x in vec1.keys()])
    sum2 = sum([vec2[x]**2 for x in vec2.keys()])
    denominator = math.sqrt(sum1) * math.sqrt(sum2)

    if not denominator:
        return 0.0
    else:
        score = float(numerator) * 100 / denominator
        return score


def compute_similarity(str1, str2, type):
    if type == 'cosine_similarity':
        return cosine_similarity(vectorize(str1), vectorize(str2.lower()))

    elif type == 'token_similarity':
        return fuzzy_token_similarity(str1, str2.lower())

    elif type == 'partial_ratio':
        return fuzzy_partial_ratio(str1, str2.lower())

    elif type == 'ratio_similarity':
        return fuzzy_ratio_similarity(str1, str2.lower())

    elif type == 'levenshtein_distance':
        return levenshtein_distance(str1, str2.lower())


def fuzzy_token_similarity(str1, str2):
    return fuzz.token_set_ratio(str1, str2)


def fuzzy_ratio_similarity(str1, str2):
    return fuzz.ratio(str1, str2)


def fuzzy_partial_ratio(str1, str2):
    return fuzz.partial_ratio(str1, str2)


def levenshtein_distance(str1, str2):
    dist = Levenshtein.distance(str1, str2)
    longer_str = str1 if len(str1) > len(str2) else str2
    score = float(dist) * 100 / len(longer_str)
    return score


def isfloat(datum):
    try:
        if datum is None or datum == '' or str(datum).isspace():
            return False
        float(datum)
    except AttributeError:
        return False
    except ValueError:
        return False
    except OverflowError:
        return False
    return True


def isint(datum):
    try:
        if datum is None or datum == '' or str(datum).isspace():
            return False
        a = float(datum)
        b = int(a)
    except AttributeError:
        return False
    except ValueError:
        return False
    except OverflowError:
        return False
    return a == b


def isdate(datum):
    try:
        if datum is None or datum == '' or str(datum).isspace():
            return False, None

        for idx, regex_list in enumerate(constants.date_regexes):
            regex = re.compile(regex_list[1])
            match = regex.match(str(datum))
            if match is not None:
                dateobj = dict()
                dateobj["regex_id"] = idx
                dateobj["regex_matches"] = list(match.groups())
                return True, dateobj

    except Exception as e:
        pass

    return False, None

# Trim the output LIST based on the debug parameter
def delete_keys_from_list(array, keys):
    keys_set = set(keys)
    modified_array = copy.deepcopy(array)
    for index, value in enumerate(array):
        if isinstance(value, MutableMapping):
            modified_array[index] = delete_keys_from_dict(value, keys_set)
        else:
            modified_array[index] = value
    return modified_array


# Trim the output DICTIONARY based on the debug parameter
def delete_keys_from_dict(dictionary, keys):
    keys_set = set(keys)
    modified_dict = {}
    for key, value in dictionary.items():
        if key not in keys_set:
            if isinstance(value, MutableMapping):
                modified_dict[key] = delete_keys_from_dict(value, keys_set)
            elif isinstance(value, list):
                modified_dict[key] = delete_keys_from_list(value, keys_set)
            else:
                modified_dict[key] = value  # or copy.deepcopy(value) if a copy is desired for non-dicts.
    return modified_dict


# Use to determine if two lists have any common elements, if any.
def common_member(a, b):
    a_set = set(a)
    b_set = set(b)
    if (a_set & b_set):
        return True
    else:
        return False


def format_str_to_date(str, format):
    try:
        return datetime.strptime(str, format)
    except Exception as e:
        return None


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, (date, datetime)):
            return obj.strftime("%Y/%m/%d")
        return json.JSONEncoder.default(self, obj)



# Map of "visualizations" that can be explicitly requested and their corresponding keywords to look for in the query.
# allow ' ' (space) in keywords? eg. bar chart / scatter plot ,etc ?
vis_keyword_map = {
    'barchart': {'bar chart', 'barchart', 'bar graph'},
    'scatterplot': {'scatter plot', 'scatterplot', 'scatter'},
    'boxplot': {'box plot', 'boxplot'},
    'linechart': {'time series', 'timeseries', 'linechart', 'line chart', 'line graph'},
    'areachart': {'area chart', 'areachart', 'area graph'},
    'stripplot': {'tick plot', 'tickplot', 'strip plot', 'stripplot'},
    'piechart': {'pie chart', 'piechart'},
    'donutchart': {'donut chart', 'donutchart'},
    'histogram': {'histogram'}
}

# Map of "tasks" that can be performed and their corresponding keywords to look for in the query.
task_keyword_map = {

    # -----------------------------------------
    # Find Extremum
    # -----------------------------------------
    "rank": [("find_extremum", "MIN")],
    "sort": [("find_extremum", "MIN")],
    "max": [("find_extremum", "MAX")],
    "maximum": [("find_extremum", "MAX")],
    "highest": [("find_extremum", "MAX")],
    "greatest": [("find_extremum", "MAX")],
    "largest": [("find_extremum", "MAX")],
    "biggest": [("find_extremum", "MAX")],
    "most": [("find_extremum", "MAX")],
    "min": [("find_extremum", "MIN")],
    "minimum": [("find_extremum", "MIN")],
    "smallest": [("find_extremum", "MIN")],
    "lowest": [("find_extremum", "MIN")],
    "least": [("find_extremum", "MIN")],  # Todo: dep parser can't detect
    "heaviest": [("find_extremum", "MAX")],
    "lightest": [("find_extremum", "MIN")],
    "best": [("find_extremum", "MAX")],
    "worst": [("find_extremum", "MIN")],
    # ADd SORT
    # -----------------------------------------

    # -----------------------------------------
    # Filter
    # -----------------------------------------
    "after": [("filter", "GT")],
    "before": [("filter", "LT")],
    "since": [("filter", "GT")],
    "until": [("filter", "LT")],
    "more": [("filter", "GT")],
    "high": [("filter", "GT")],
    "over": [("filter", "GT")],
    "higher": [("filter", "GT")],
    "greater": [("filter", "GT")],
    "larger": [("filter", "GT")],
    "bigger": [("filter", "GT")],
    "under": [("filter", "LT")],
    "less": [("filter", "LT")],
    "lower": [("filter", "LT")],
    "lesser": [("filter", "LT")],
    "smaller": [("filter", "LT")],
    "between": [("filter", "RANGE")],
    # -----------------------------------------

    # -----------------------------------------
    # Compute Derived Value
    # -----------------------------------------
    "median": [("derived_value", "MEDIAN")],
    "average": [("derived_value", "AVG")],
    "mean": [("derived_value", "AVG")],
    "sum": [("derived_value", "SUM")],
    "total": [("derived_value", "SUM")],
    # -----------------------------------------

    # -----------------------------------------
    # Characterize Distribution
    # -----------------------------------------
    "distribute": [("distribution", None)],
    "distribution": [("distribution", None)],
    "range": [("distribution", None)],
    "extent": [("distribution", None)],
    # -----------------------------------------

    # -----------------------------------------
    # Outliers / Anomalies
    # -----------------------------------------
    "outlier": [("outlier",)],
    "exception": [("outlier",)],
    # -----------------------------------------

    # -----------------------------------------
    # Correlation
    # -----------------------------------------
    "correlation": [("correlation", None)],
    "correlate": [("correlation", None)],
    "relationship": [("correlation", None)],
    # -----------------------------------------
    # Negation
    # -----------------------------------------
    "not": [("negation", None)],

    # -----------------------------------------
    # Equality
    # -----------------------------------------
    "equals": [("filter", "EQ")],
    "equal": [("filter", "EQ")],

    # -----------------------------------------
    # Trend
    # -----------------------------------------
    "trend": [("trend", None)]
}

# scores given to attribute/ task/ vis matches of different types
match_scores = {
    'attribute': {
        'attribute_exact_match': 1,
        'attribute_similarity_match': 0.9,
        'attribute_alias_exact_match': 0.8,
        'attribute_alias_similarity_match': 0.75,
        'attribute_synonym_match': 0.5,
        'attribute_domain_value_match': 0.5,
    },
    'task': {
        'explicit': 1,
        'implicit': 0.5,
    },
    'vis': {
        'explicit': 1
    }
}

# JSON has an abbreviated version (GT) but Vega-Lite requires symbols (>)
operator_symbol_mapping = {
    "NOT RANGE": "not between",
    "GT": ">",
    "LT": "<",
    "EQ": "==",
    "SUM": "sum",
    "MEDIAN": "median",
    "AVG": "mean",
    "IN": "in",
    "RANGE": "between",
    "MAX": "max",
    "MIN": "min"
}

# keep them high, so single letter typos in shorter words are not caught. eg. with > width
match_thresholds = {
    'synonymity': 95,
    'string_similarity': 85
}

# These keys are deleted from final output since they are NOT important/critical. They are retained when the `debug` flag is true.
keys_to_delete_in_output = ['matchScore',
                            'meta',
                            'debug',
                            'score',
                            'status',
                            'scoreObj',
                            'encode',
                            'isLabel',
                            'metric',
                            'isAttrAmbiguous',
                            'isValueAmbiguous',
                            'confidenceObj']

attribute_types = {
    'QUANTITATIVE': 'Q',
    'NOMINAL': 'N',
    'TEMPORAL': 'T',
    'ORDINAL': 'O',
}

# NL4DV datatypes to VL datatypes mapping
vl_attribute_types = {
    'Q': 'quantitative',
    'N': 'nominal',
    'T': 'temporal',
    'O': 'ordinal',
}

# Regular Expressions with their corresponding FORMATs and Examples
date_regexes = [
    # Format:
        # MM*DD*YY(YY) where * ∈ {. - /}
    # Examples:
        # 12.24.2019
        # 12:24:2019
        # 12-24-2019
        # 12/24/2019
        # 1/24/2019
        # 07/24/2019
        # 1/24/20
    [['%m/%d/%Y', '%m/%d/%y'], "([1][0-2]|[0]?[1-9])[-.\/]([1|2][0-9]|[3][0|1]|[0]?[1-9])[-.\/]([1-9][0-9]{3}|[0-9]{2})"],
    # Format:
        # YY(YY)*MM*DD where * ∈ {. - /}
    # Examples:
        # 2019.12.24
        # 2019.12.24
        # 2019-12-24
        # 2019/12/24
        # 2019/1/24
        # 2019/07/24
        # 20/1/24
    [['%Y/%m/%d', '%y/%m/%d'], "([1-9][0-9]{3}|[0-9]{2})[-.\/]([1][0-2]|[0]?[1-9])[-.\/]([1|2][0-9]|[3][0|1]|[0]?[1-9])"],
    # Format:
        # DD*MM*YY(YY) where * ∈ {. - /}
    # Examples:
        # 24.12.2019
        # 24:12:2019
        # 24-12-2019
        # 24/12/2019
        # 24/1/2019
        # 24/07/2019
        # 24/1/20
    [['%d/%m/%Y', '%d/%m/%y'], "([1|2][0-9]|[3][0|1]|[0]?[1-9])[-.\/]([1][0-2]|[0]?[1-9])[-.\/]([1-9][0-9]{3}|[0-9]{2})"],
    # Formats:
        # DD*MMM(M)*YY(YY) where * ∈ {. - / <space>}
    # Examples:
        # 8-January-2019
        # 31 Dec 19
    [['%d/%b/%Y', '%d/%B/%Y', '%d/%b/%y', '%d/%B/%y'], "([1|2][0-9]|[3][0|1]|[0]?[1-9])[-.\/\s](January|February|March|April|May|June|July|August|September|October|November|December|Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[-.\/\s]([1-9][0-9]{3}|[0-9]{2})"],
    # Format:
        # DD*MMM(M) where * ∈ {. - / <space>}
    # Examples:
        # 31-January
        # 1 Jul
    [['%d/%b', '%d/%B'], "([1|2][0-9]|[3][0|1]|[0]?[1-9])[-.\/\s](January|February|March|April|May|June|July|August|September|October|November|December|Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)"],
    # Formats:
        # MMM(M)*DD*YYY(Y) where * ∈ {. - / <space>}
    # Examples:
        # January-8-2019
        # Dec 31 19
    [['%b/%d/%Y', '%B/%d/%Y', '%b/%d/%y', '%B/%d/%y'], "(January|February|March|April|May|June|July|August|September|October|November|December|Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[-.\/\s]([1|2][0-9]|[3][0|1]|[0]?[1-9])[-.\/\s]([1-9][0-9]{3}|[0-9]{2})"],
    # Format:
        # MMM(M)*DD where * ∈ {. - / <space>}
    # Examples:
        # January-31
        # Jul 1
    [['%b/%d', '%B/%d'], "(January|February|March|April|May|June|July|August|September|October|November|December|Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)[-.\/\s]([1|2][0-9]|[3][0|1]|[0]?[1-9])"],
    # Format:
        # YYYY
    # Examples:
        # 18XX, 19XX, 20XX
    [["%Y"], "(1[8-9][0-9][0-9]|20[0-2][0-9])"]
]



from nl4dv.utils import constants, helpers
import re
import nltk
from nltk import word_tokenize
from si_prefix import si_parse
from nltk.corpus import stopwords


class QueryGenie:
    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance

        # Stopwords
        self.stopwords_set = set(stopwords.words('english'))

    def process_query(self, query):

        # Try to infer and convert numerical shorthands to machine understandable format, e.g. 100M = 100*1,000,000
        parsed_tokens = []
        for token in word_tokenize(query):
            try:
                parsed_tokens.append(str(int(si_parse(token))))
            except Exception as e:
                parsed_tokens.append(token)

        # Convert to lowercase
        query_lower = ' '.join(parsed_tokens).lower()

        return query_lower

    def clean_query_and_get_query_tokens(self, query, reserve_words, ignore_words):

        # Clean sentence of non-alphanumerical characters
        query_alphanumeric = re.sub(r'[^A-Za-z0-9]+', ' ', query)

        # Set the stopwords from reserve words and ignore words
        self.stopwords_set = self.stopwords_set.difference(set(reserve_words)).union(set(ignore_words))

        # Create token set and filter out standard stopwords
        query_tokens = list(filter(lambda token: token not in self.stopwords_set, word_tokenize(query_alphanumeric)))

        return query_tokens

    # 1-time generation of query N-GRAMS
    def get_query_ngrams(self, query):
        # Do NOT check for n_grams with numeric entities in the domain. They tend to produce erroneous results, especially due to the TOKEN based similarity algorithm.
        query_alpha_str = ''.join([i for i in query if not i.isdigit()])
        ngrams = dict()
        for i in range(len(query_alpha_str.split()), 0, -1):
            for ngram in helpers.get_ngrams(query_alpha_str, i):
                ngram_str = ((' '.join(map(str, ngram))).rstrip()).lower()
                ngrams[ngram_str] = dict()
                ngrams[ngram_str]['raw'] = ngram
                ngrams[ngram_str]['lower'] = ngram_str
                ngrams[ngram_str]['stemmed_lower'] = ' '.join(self.nl4dv_instance.porter_stemmer_instance.stem(t) for t in nltk.word_tokenize(ngram_str))

        return ngrams

    # Create a parts of speech map from the Query Tokens
    def get_query_pos_map(self, query_tokens):
        pos_map = dict()
        pos_list = nltk.pos_tag(query_tokens)
        for pos in pos_list:
            pos_map[pos[0]] = pos[1]



from collections import Counter
import json
import sys
from nl4dv.utils import constants, error_codes, helpers
import os
import pandas as pd
import requests

class DataGenie:
    """
    Pre-process data attributes into different categories
    """

    def __init__(self, nl4dv_instance):

        # nl4dv instance
        self.nl4dv_instance = nl4dv_instance
        
        # set label attribute
        self.set_label_attribute(label_attribute=self.nl4dv_instance.label_attribute)

        # set ignore list
        self.set_ignore_words(ignore_words=self.nl4dv_instance.ignore_words)

        # set custom stopwords list
        self.set_reserve_words(reserve_words=self.nl4dv_instance.reserve_words)

        # Other initializations
        self.data_attribute_map = dict()
        self.data = list()
        self.rows = 0

        # Set the Data if passed data_url or data_value is not None
        if self.nl4dv_instance.data_url is not None:
            self.set_data(data_url = self.nl4dv_instance.data_url)
        elif self.nl4dv_instance.data_value is not None:
            self.set_data(data_value = self.nl4dv_instance.data_value)

        # Set the Aliases if passed alias_url or alias_value is not None
        if self.nl4dv_instance.alias_url is not None:
            self.set_alias_map(alias_url = self.nl4dv_instance.alias_url)
        elif self.nl4dv_instance.alias_value is not None:
            self.set_alias_map(alias_value = self.nl4dv_instance.alias_value)

    # Update the attribute datatypes that were not correctly detected by NL4DV
    def set_attribute_datatype(self, attr_type_obj):
        # Set new datatype
        for attribute, data_type in attr_type_obj.items():
            if data_type in constants.attribute_types.values():
                self.data_attribute_map[attribute]['dataType'] = data_type
                self.populate_dataset_meta_for_attr(attribute, data_type)
            else:
                helpers.cond_print("Invalid Target DataType. Choose from " + str(constants.attribute_types.values()), debug=True)
                sys.exit(error_codes.BAD_INPUT_ATTRIBUTE_DATA_TYPE)

    # Set Label attribute for the dataset, i.e. one that defines what the dataset is about.
    # e.g. "Correlate horsepower and MPG for sports car models" should NOT apply an explicit attribute for models since there are two explicit attributes already present.
    def set_label_attribute(self, label_attribute):
        self.nl4dv_instance.label_attribute = label_attribute

    # WORDS that should be IGNORED in the query, i.e. NOT lead to the detection of attributes and tasks
    # `Movie` in movies dataset
    # `Car` in cars dataset
    def set_ignore_words(self, ignore_words):
        self.nl4dv_instance.ignore_words = ignore_words

    # Custom STOPWORDS that should NOT removed from the query, as they might be present in the domain.
    # e.g. `A` in grades dataset
    def set_reserve_words(self, reserve_words):
        self.nl4dv_instance.reserve_words = reserve_words

    # Sets the Dataset
    def set_data(self, data_url=None, data_value=None):
        # type: (str) -> None
        """
        User can choose to manually initialize data

        """
        self.nl4dv_instance.data_url = data_url if data_url is not None else self.nl4dv_instance.data_url
        self.nl4dv_instance.data_value = data_value if data_value is not None else self.nl4dv_instance.data_value
        self.nl4dv_instance.data_url_type = None

        # initialize values
        self.data_attribute_map = dict()
        self.data = list()
        self.rows = 0

        if self.nl4dv_instance.data_url is not None:
            # Possible Local FILE or HTTP URL
            if self.nl4dv_instance.data_url.lower().endswith('.csv'):
                self.nl4dv_instance.data_url_type = "csv"
                self.data = pd.read_csv(self.nl4dv_instance.data_url, sep=',').to_dict('records')
            elif self.nl4dv_instance.data_url.lower().endswith('.tsv'):
                self.nl4dv_instance.data_url_type = "tsv"
                self.data = pd.read_csv(self.nl4dv_instance.data_url, sep='\t').to_dict('records')
            elif self.nl4dv_instance.data_url.lower().endswith('.json'):
                self.nl4dv_instance.data_url_type = "json"
                self.data = pd.read_json(self.nl4dv_instance.data_url).to_dict('records')

        elif self.nl4dv_instance.data_value is not None:
            if isinstance(data_value, pd.DataFrame):
                self.data = data_value.to_dict('records')
            elif isinstance(data_value, list):
                self.data = data_value
            elif isinstance(data_value, dict):
                self.data = pd.DataFrame(data_value).to_dict('records')

        # Set number of rows in the dataset
        self.rows = len(self.data)

        # initialize properties in Attribute Map
        if len(self.data) > 0:
            for attr in self.data[0].keys():
                # Don't consider attribute names that are empty or just whitespaces
                if attr and attr.strip():
                    self.data_attribute_map[attr] = {
                        'domain': set(),
                        'domainMeta': dict(),
                        'isLabelAttribute': attr == self.nl4dv_instance.label_attribute,
                        'summary': dict(),
                        'dataTypeList': list(), # temporary to determine datatype
                        'dataType': '',
                        'dataTypeMeta': dict(),  # Used for for e.g., temporal attributes when they conform to a certain format
                        'aliases': list(),
                    }

        # infer attribute datatypes and compute summary (range, domain)
        for datum in self.data:
            for attr in self.data_attribute_map.keys():
                # Check for Datetime
                is_date, unformatted_date_obj = helpers.isdate(datum[attr])
                if is_date:
                    attr_datatype_for_majority = constants.attribute_types['TEMPORAL'] + "-" + str(unformatted_date_obj["regex_id"])
                # Check for Numeric (float, int)
                elif helpers.isfloat(datum[attr]) or helpers.isint(datum[attr]):
                    attr_datatype_for_majority = constants.attribute_types['QUANTITATIVE']
                # Otherwise set as Nominal
                else:
                    attr_datatype_for_majority = constants.attribute_types['NOMINAL']

                # Append the list of attribute types for each data row to take best decision on heterogeneous data with multiple datatypes
                self.data_attribute_map[attr]['dataTypeList'].append(attr_datatype_for_majority)

        # Determine the Datatype based on majority of values.
        # Also Override a few datatypes set above based on rules such as NOMINAL to ORDINAL if all values are unique such as Sr. 1, Sr. 2, ...
        for attr in self.data_attribute_map:
            # By default, set the attribute datatype to the most common attribute
            attr_datatype = Counter(self.data_attribute_map[attr]['dataTypeList']).most_common(1)[0][0]

            # If attr_datatype is Temporal (e.g., T-1, T-2, T-n where 'n' corresponds to the n'th index of the date_regex array.
            # Then: PROCESS this and eventually strip the '-n' from the datatype
            if not (attr_datatype in [constants.attribute_types['QUANTITATIVE'], constants.attribute_types['NOMINAL']]):

                # If there is at least one instance of 'T-2' (DD*MM*YY(YY)), in the `dataTypeList`, set the regex_id to this, even if 'T-1' is the majority.
                if 'T-2' in self.data_attribute_map[attr]['dataTypeList']:
                    attr_datatype = 'T-2'

                # Strip the '-n' from the datatype
                attr_datatype_split = attr_datatype.split("-")
                attr_datatype = attr_datatype_split[0]

                # Set the final data type
                self.data_attribute_map[attr]['dataTypeMeta'] = {
                    "regex_id": attr_datatype_split[1]
                }

                # Add raw data to the domain's metadata. Only for Temporal Attributes.
                if not 'raw' in self.data_attribute_map[attr]['domainMeta']:
                    self.data_attribute_map[attr]['domainMeta']['raw'] = set()

            # Set the final data type
            self.data_attribute_map[attr]['dataType'] = attr_datatype

            # Update the dataset metadata for each attribute
            self.populate_dataset_meta_for_attr(attr, attr_datatype)

    # Sets the Alias Map
    def set_alias_map(self, alias_value=None, alias_url=None):
        # type: (dict, str) -> None
        """
        User can choose to manually initialize aliases

        """
        self.nl4dv_instance.alias_url= alias_url
        self.nl4dv_instance.alias_value = alias_value

        if self.nl4dv_instance.alias_url is not None:
            if os.path.isfile(self.nl4dv_instance.alias_url):
                self.nl4dv_instance.alias_value = json.load(open(self.nl4dv_instance.alias_url, 'r', encoding='utf-8'))
            else:
                self.nl4dv_instance.alias_value = json.loads(requests.get(self.nl4dv_instance.alias_url).text)

        if self.nl4dv_instance.alias_value is not None:
            for attr in self.nl4dv_instance.alias_value.keys():
                if attr in self.data_attribute_map:
                    self.data_attribute_map[attr]['aliases'].extend(self.nl4dv_instance.alias_value[attr])

    # Set Domain, Ranges for Attributes of different datatypes
    def populate_dataset_meta(self, attr, attr_val, attr_datatype):
        if attr_datatype == constants.attribute_types['QUANTITATIVE']:
            try:
                attr_val = float(attr_val)
            except Exception as e:
                attr_val = float('NaN')
            self.data_attribute_map[attr]['domain'].add(attr_val)

            # Compute Max and Min of the attribute values
            if 'min' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['min'] = float("inf")
            if 'max' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['max'] = float("-inf")

            if attr_val > self.data_attribute_map[attr]['summary']['max']:
                self.data_attribute_map[attr]['summary']['max'] = attr_val
            if attr_val < self.data_attribute_map[attr]['summary']['min']:
                self.data_attribute_map[attr]['summary']['min'] = attr_val

        elif attr_datatype == constants.attribute_types['TEMPORAL']:
            is_date, unformatted_date_obj = helpers.isdate(attr_val)
            parsed_attr_val = None
            if is_date:
                for format in constants.date_regexes[unformatted_date_obj['regex_id']][0]:
                    parsed_attr_val = helpers.format_str_to_date("/".join(unformatted_date_obj["regex_matches"]), format)
                    if parsed_attr_val is not None:
                        self.data_attribute_map[attr]['domain'].add(parsed_attr_val)
                        self.data_attribute_map[attr]['domainMeta']['raw'].add(attr_val)
                        break

            if parsed_attr_val is not None:
                # Compute Max and Min of the attribute datetime
                if 'start' not in self.data_attribute_map[attr]['summary']:
                    self.data_attribute_map[attr]['summary']['start'] = parsed_attr_val

                if 'end' not in self.data_attribute_map[attr]['summary']:
                    self.data_attribute_map[attr]['summary']['end'] = parsed_attr_val

                # print(parsed_status, attr_val, parsed_attr_val, self.data_attribute_map[attr]['summary']['end'])
                if parsed_attr_val > self.data_attribute_map[attr]['summary']['end']:
                    self.data_attribute_map[attr]['summary']['end'] = parsed_attr_val

                if parsed_attr_val < self.data_attribute_map[attr]['summary']['start']:
                    self.data_attribute_map[attr]['summary']['start'] = parsed_attr_val

        else:
            attr_val = str(attr_val)
            self.data_attribute_map[attr]['domain'].add(attr_val)
            # Compute # occurrences of attribute values
            if 'group_counts' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['group_counts'] = dict()
            if attr_val not in self.data_attribute_map[attr]['summary']['group_counts']:
                self.data_attribute_map[attr]['summary']['group_counts'][attr_val] = 0
            self.data_attribute_map[attr]['summary']['group_counts'][attr_val] += 1

    # Set Domain, Ranges for single attribute
    def populate_dataset_meta_for_attr(self, attr, attr_datatype):
        # Re-initialize the dataTypeList to re-set the domain based on the new Prototype
        self.data_attribute_map[attr]['dataTypeList'] = list()

        # For each value in the dataset, add to the attribute map (domain, range)
        self.data_attribute_map[attr]['domain'] = set()
        for datum in self.data:
            self.populate_dataset_meta(attr, datum[attr], attr_datatype)

        # If an attribute has (almost) no repeating value, then mark it as the label attribute.
        # eg. primary/unique key of the table? Car1 , Car2, Car3, ...
        # Almost == 90% (heuristic-based)
        if attr_datatype == constants.attribute_types['NOMINAL'] and len(self.data_attribute_map[attr]['domain']) > 0.9 * self.rows:
            self.nl4dv_instance.label_attribute = attr
            self.data_attribute_map[attr]['isLabelAttribute'] = True

        # Sort domain values, remove unwanted keys, etc.
        self.format_data_attribute_map(attr, attr_datatype)

    def format_data_attribute_map(self, attr, attr_datatype):
        # Remove NANs which are supposedly NOT unique
        self.data_attribute_map[attr]['domain'] = set(filter(lambda x: x == x, self.data_attribute_map[attr]['domain']))

        # Set > Sorted List > Set
        try:
            if attr_datatype == constants.attribute_types['QUANTITATIVE']:
                self.data_attribute_map[attr]['domain'] = set(sorted([float(a) for a in self.data_attribute_map[attr]['domain']]))
            elif attr_datatype == constants.attribute_types['TEMPORAL']:
                self.data_attribute_map[attr]['domain'] = set(sorted(self.data_attribute_map[attr]['domain']))
                self.data_attribute_map[attr]['domainMeta']['raw'] = set(sorted(self.data_attribute_map[attr]['domainMeta']['raw']))
            else:
                self.data_attribute_map[attr]['domain'] = set(sorted(self.data_attribute_map[attr]['domain']))
        except Exception as e:
            # works for hybrid numbers + strings combinations
            self.data_attribute_map[attr]['domain'] = set(sorted([str(a) for a in self.data_attribute_map[attr]['domain']]))

        # Delete the keys that are now redundant
        self.delete_unwanted_keys(attr, attr_datatype)

    # Generic method to delete keys based on datatype
    def delete_unwanted_keys(self, attr, attr_datatype):
        # Delete the list of datatypes, not needed anymore
        if 'dataTypeList' in self.data_attribute_map[attr]:
            del self.data_attribute_map[attr]['dataTypeList']

        if attr_datatype == constants.attribute_types['NOMINAL'] or attr_datatype == constants.attribute_types['ORDINAL']:
            self.del_temporal_keys(attr)
            self.del_quantitative_keys(attr)

        elif attr_datatype == constants.attribute_types['QUANTITATIVE']:
            self.del_temporal_keys(attr)
            self.del_nominal_keys(attr)

        elif attr_datatype == constants.attribute_types['TEMPORAL']:
            self.del_quantitative_keys(attr)
            self.del_nominal_keys(attr)

    # Delete Keys from attribute that is determined to be Temporal
    def del_temporal_keys(self, attr):
        if 'start' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['start']
        if 'end' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['end']

    # Delete Keys from attribute that is determined to be Quantitative
    def del_quantitative_keys(self, attr):
        # Keys from Quantitative
        if 'min' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['min']
        if 'max' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['max']

    # Delete Keys from attribute that is determined to be Nominal
    def del_nominal_keys(self, attr):
        # Keys from Nominal
        if 'group_counts' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['group_counts']



vis_design_combos = {
    'Q': {
        'combination': 'Q',
        'support': True,
        'tasks': ['distribution'],
        'visualizations': ['histogram', 'stripplot', 'boxplot'],
        'designs': [{
            'task': 'distribution',
            'vis_type': 'histogram',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y'],
            'priority': ['x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x'],
            'priority': ['x'],
            'x': {'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'boxplot',
            'not_suggested_by_default': False,
            'mark': 'boxplot',
            'mandatory': ['x'],
            'priority': ['x'],
            'x': {'is_defined': False, 'agg': None}
        }]
    },
    'N': {
        'combination': 'N',
        'support': True,
        'tasks': ['distribution'],
        'visualizations': ['barchart'],
        'designs': [{
            'task': 'distribution',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y'],
            'priority': ['x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        }]
    },
    'O': {
        'combination': 'O',
        'support': True,
        'tasks': ['distribution'],
        'visualizations': ['barchart'],
        'designs': [{
            'task': 'distribution',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y'],
            'priority': ['x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        }]
    },
    'T': {
        'combination': 'T',
        'support': True,
        'tasks': ['trend'],
        'visualizations': ['linechart', 'areachart'],
        'designs': [{
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y'],
            'priority': ['x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        },{
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y'],
            'priority': ['x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        }]
    },
    'QQ': {
        'combination': 'QQ',
        'support': True,
        'tasks': ['correlation'],
        'visualizations': ['scatterplot'],
        'designs': [{
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y'],
            'priority': ['x', 'y'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QN': {
        'combination': 'QN',
        'support': True,
        'tasks': ['derived_value', 'distribution'],
        'visualizations': ['barchart', 'piechart', 'donutchart', 'tickplot'],
        'designs': [{
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y'],
            'priority': ['y', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'derived_value',
            'vis_type': 'piechart',
            'not_suggested_by_default': True,
            'mark': 'arc',
            'mandatory': ['theta', 'color'],
            'priority': ['theta', 'color'],
            'color': {'attr': None, 'is_defined': False, 'agg': None},
            'theta': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'derived_value',
            'vis_type': 'donutchart',
            'not_suggested_by_default': True,
            'mark': 'arc',
            'mandatory': ['theta', 'color'],
            'priority': ['theta', 'color'],
            'color': {'attr': None, 'is_defined': False, 'agg': None},
            'theta': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y'],
            'priority': ['y', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QO': {
        'combination': 'QO',
        'support': True,
        'tasks': ['derived_value', 'distribution'],
        'visualizations': ['barchart', 'piechart', 'donutchart', 'tickplot'],
        'designs': [{
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y'],
            'priority': ['y', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'derived_value',
            'vis_type': 'piechart',
            'not_suggested_by_default': True,
            'mark': 'arc',
            'mandatory': ['theta', 'color'],
            'priority': ['theta', 'color'],
            'color': {'attr': None, 'is_defined': False, 'agg': None},
            'theta': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'derived_value',
            'vis_type': 'donutchart',
            'not_suggested_by_default': True,
            'mark': 'arc',
            'mandatory': ['theta', 'color'],
            'priority': ['theta', 'color'],
            'color': {'attr': None, 'is_defined': False, 'agg': None},
            'theta': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y'],
            'priority': ['y', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QT': {
        'combination': 'QT',
        'support': True,
        'tasks': ['trend', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'stripplot'],
        'designs': [{
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y'],
            'priority': ['y', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y'],
            'priority': ['y', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y'],
            'priority': ['y', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'NN': {
        'combination': 'NN',
        'support': True,
        'tasks': ['distribution'],
        'visualizations': ['scatterplot', 'barchart'],
        'designs': [{
            'task': 'distribution',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['x', 'y'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        }, {
            'task': 'distribution',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'NO': {
        'combination': 'NO',
        'support': True,
        'tasks': ['distribution'],
        'visualizations': ['scatterplot', 'barchart'],
        'designs': [{
            'task': 'distribution',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['x', 'y'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        }, {
            'task': 'distribution',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'NT': {
        'combination': 'NT',
        'support': True,
        'tasks': ['trend', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'stripplot'],
        'designs': [{
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }]
    },
    'OO': {
        'combination': 'OO',
        'support': True,
        'tasks': ['distribution'],
        'visualizations': ['scatterplot', 'barchart'],
        'designs': [{
            'task': 'distribution',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['x', 'y'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'}
        }, {
            'task': 'distribution',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'OT': {
        'combination': 'OT',
        'support': True,
        'tasks': ['trend', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'stripplot'],
        'designs': [{
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }]
    },
    'TT': {
        'combination': 'TT',
        'support': False,
        'tasks': [],
        'visualizations': [],
        'designs': []
    },
    'QQQ': {
        'combination': 'QQQ',
        'support': True,
        'tasks': ['correlation'],
        'visualizations': ['scatterplot'],
        'designs': [{
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['x', 'y', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['x', 'y', 'size'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QQN': {
        'combination': 'QQN',
        'support': True,
        'tasks': ['correlation', 'distribution'],
        'visualizations': ['scatterplot', 'stripplot'],
        'designs': [{
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['x', 'y', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['x', 'y', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QQO': {
        'combination': 'QQO',
        'support': True,
        'tasks': ['correlation', 'distribution'],
        'visualizations': ['scatterplot', 'stripplot'],
        'designs': [{
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['x', 'y', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['x', 'y', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QQT': {
        'combination': 'QQT',
        'support': True,
        'tasks': ['correlation'],
        'visualizations': ['scatterplot'],
        'designs': [{
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['x', 'y', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'correlation',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['x', 'y', 'size'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QNN': {
        'combination': 'QNN',
        'support': True,
        'tasks': ['derived_value', 'distribution'],
        'visualizations': ['barchart', 'scatterplot', 'stripplot'],
        'designs': [{
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'sum'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['size', 'x', 'y'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'x', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['y', 'x', 'size'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QNO': {
        'combination': 'QNO',
        'support': True,
        'tasks': ['derived_value', 'distribution'],
        'visualizations': ['barchart', 'scatterplot', 'stripplot'],
        'designs': [{
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'sum'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['size', 'x', 'y'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'x', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['y', 'x', 'size'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QNT': {
        'combination': 'QNT',
        'support': True,
        'tasks': ['derived_value', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'barchart', 'stripplot'],
        'designs': [{
            'task': 'derived_value',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'x', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'x', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': None}
        }]
    },
    'QOO': {
        'combination': 'QOO',
        'support': True,
        'tasks': ['derived_value', 'distribution'],
        'visualizations': ['barchart', 'scatterplot', 'stripplot'],
        'designs': [{
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'sum'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['size', 'x', 'y'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': 'mean'}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'x', 'color'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'x', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'size'],
            'priority': ['y', 'x', 'size'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, 'agg': None}
        }]
    },
    'QOT': {
        'combination': 'QOT',
        'support': True,
        'tasks': ['derived_value', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'barchart', 'stripplot'],
        'designs': [{
            'task': 'derived_value',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'derived_value',
            'vis_type': 'barchart',
            'not_suggested_by_default': False,
            'mark': 'bar',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'x', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': 'mean'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'column', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'column', 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'color'],
            'priority': ['y', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'stripplot',
            'not_suggested_by_default': False,
            'mark': 'tick',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['y', 'x', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': None}
        }]
    },
    'QTT': {
        'combination': 'QTT',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    },
    'NNN': {
        'combination': 'NNN',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    },
    'NNO': {
        'combination': 'NNO',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    },
    'NNT': {
        'combination': 'NNT',
        'support': True,
        'tasks': ['trend', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'scatterplot'],
        'designs': [{
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'column', 'size'],
            'priority': ['x', 'y', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, "attr_ref": "x", "agg": "count"}
        }]
    },
    'NOO': {
        'combination': 'NOO',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    },
    'NOT': {
        'combination': 'NOT',
        'support': True,
        'tasks': ['trend', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'scatterplot'],
        'designs': [{
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'column', 'size'],
            'priority': ['x', 'y', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, "attr_ref": "x", "agg": "count"}
        }]
    },
    'NTT': {
        'combination': 'NTT',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    },
    'OOO': {
        'combination': 'OOO',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    },
    'OOT': {
        'combination': 'OOT',
        'support': True,
        'tasks': ['trend', 'distribution'],
        'visualizations': ['linechart', 'areachart', 'scatterplot'],
        'designs': [{
            'task': 'trend',
            'vis_type': 'linechart',
            'not_suggested_by_default': False,
            'mark': 'line',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'trend',
            'vis_type': 'areachart',
            'not_suggested_by_default': True,
            'mark': 'area',
            'mandatory': ['x', 'y', 'column', 'color'],
            'priority': ['column', 'color', 'x'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'attr_ref': 'x', 'agg': 'count'},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'color': {'attr': None, 'is_defined': False, 'agg': None}
        }, {
            'task': 'distribution',
            'vis_type': 'scatterplot',
            'not_suggested_by_default': False,
            'mark': 'point',
            'mandatory': ['x', 'y', 'column', 'size'],
            'priority': ['x', 'y', 'column'],
            'x': {'attr': None, 'is_defined': False, 'agg': None},
            'y': {'attr': None, 'is_defined': False, 'agg': None},
            'column': {'attr': None, 'is_defined': False, 'agg': None},
            'size': {'attr': None, 'is_defined': False, "attr_ref": "x", "agg": "count"}
        }]
    },
    'OTT': {
        'combination': 'OTT',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    },
    'TTT': {
        'combination': 'TTT',
        'support': False,
        'designs': [],
        'tasks': [],
        'visualizations': []
    }
}
